@@ -12,9 +12,9 @@ use std::{
 use super::{
     super::{
         base::{self, PortDeserializer, PortSerializer},
-        buffer, RemoteSendError, BACKCHANNEL_MSG_CLOSE, BACKCHANNEL_MSG_ERROR,
+        buffer, ClosedReason, RemoteSendError, BACKCHANNEL_MSG_CLOSE, BACKCHANNEL_MSG_ERROR,
     },
-    Distributor,
+    recv_impl, send_impl, Distributor,
 };
 use crate::{chmux, codec, RemoteSend};
 
@@ -51,6 +51,57 @@ impl RecvError {
     }
 }
 
+/// An error occurred during trying to receive over an mpsc channel.
+#[derive(Clone, Debug, Serialize, Deserialize)]
+pub enum TryRecvError {
+    /// All channel senders have been dropped.
+    Closed,
+    /// Currently no value is ready to receive, but values may still arrive
+    /// in the future.
+    Empty,
+    /// Receiving from a remote endpoint failed.
+    RemoteReceive(base::RecvError),
+    /// Connecting a sent channel failed.
+    RemoteConnect(chmux::ConnectError),
+    /// Listening for a connection from a received channel failed.
+    RemoteListen(chmux::ListenerError),
+}
+
+impl fmt::Display for TryRecvError {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::Closed => write!(f, "channel is closed"),
+            Self::Empty => write!(f, "channel is empty"),
+            Self::RemoteReceive(err) => write!(f, "receive error: {}", err),
+            Self::RemoteConnect(err) => write!(f, "connect error: {}", err),
+            Self::RemoteListen(err) => write!(f, "listen error: {}", err),
+        }
+    }
+}
+
+impl From<RecvError> for TryRecvError {
+    fn from(err: RecvError) -> Self {
+        match err {
+            RecvError::RemoteReceive(err) => Self::RemoteReceive(err),
+            RecvError::RemoteConnect(err) => Self::RemoteConnect(err),
+            RecvError::RemoteListen(err) => Self::RemoteListen(err),
+        }
+    }
+}
+
+impl Error for TryRecvError {}
+
+impl TryRecvError {
+    /// Returns whether the error is final, i.e. no further receive operation can succeed.
+    pub fn is_final(&self) -> bool {
+        match self {
+            Self::Empty => false,
+            Self::RemoteReceive(err) => err.is_final(),
+            Self::Closed | Self::RemoteConnect(_) | Self::RemoteListen(_) => true,
+        }
+    }
+}
+
 /// Receive values from the associated [Sender](super::Sender),
 /// which may be located on a remote endpoint.
 ///
@@ -59,6 +110,7 @@ pub struct Receiver<T, Codec = codec::Default, Buffer = buffer::Default> {
     inner: Option<ReceiverInner<T>>,
     #[allow(clippy::type_complexity)]
     successor_tx: Mutex<Option<tokio::sync::oneshot::Sender<ReceiverInner<T>>>>,
+    final_err: Option<RecvError>,
     _codec: PhantomData<Codec>,
     _buffer: PhantomData<Buffer>,
 }
@@ -71,8 +123,9 @@ impl<T, Codec, Buffer> fmt::Debug for Receiver<T, Codec, Buffer> {
 
 pub(crate) struct ReceiverInner<T> {
     rx: tokio::sync::mpsc::Receiver<Result<T, RecvError>>,
-    closed_tx: tokio::sync::watch::Sender<bool>,
+    closed_tx: tokio::sync::watch::Sender<Option<ClosedReason>>,
     remote_send_err_tx: tokio::sync::watch::Sender<Option<RemoteSendError>>,
+    closed: bool,
 }
 
 /// Mpsc receiver in transport.
@@ -84,16 +137,21 @@ pub(crate) struct TransportedReceiver<T, Codec> {
     data: PhantomData<T>,
     /// Data codec.
     codec: PhantomData<Codec>,
+    /// Receiver has been closed.
+    #[serde(default)]
+    closed: bool,
 }
 
 impl<T, Codec, Buffer> Receiver<T, Codec, Buffer> {
     pub(crate) fn new(
-        rx: tokio::sync::mpsc::Receiver<Result<T, RecvError>>, closed_tx: tokio::sync::watch::Sender<bool>,
+        rx: tokio::sync::mpsc::Receiver<Result<T, RecvError>>,
+        closed_tx: tokio::sync::watch::Sender<Option<ClosedReason>>, closed: bool,
         remote_send_err_tx: tokio::sync::watch::Sender<Option<RemoteSendError>>,
     ) -> Self {
         Self {
-            inner: Some(ReceiverInner { rx, closed_tx, remote_send_err_tx }),
+            inner: Some(ReceiverInner { rx, closed_tx, remote_send_err_tx, closed }),
             successor_tx: Mutex::new(None),
+            final_err: None,
             _codec: PhantomData,
             _buffer: PhantomData,
         }
@@ -101,25 +159,93 @@ impl<T, Codec, Buffer> Receiver<T, Codec, Buffer> {
 
     /// Receives the next value for this receiver.
     ///
-    /// This function returns `Ok(None)` when the channel sender has been dropped.
+    /// This function returns `Ok(None)` when all channel senders have been dropped.
+    ///
+    /// When a receive error occurs due to a connection failure and other senders are still
+    /// present, it is held back and returned after all other senders have been dropped or failed.
+    /// Use [error](Self::error) to check if such an error is present.
     #[inline]
     pub async fn recv(&mut self) -> Result<Option<T>, RecvError> {
-        match self.inner.as_mut().unwrap().rx.recv().await {
-            Some(Ok(value_opt)) => Ok(Some(value_opt)),
-            Some(Err(err)) => Err(err),
-            None => Ok(None),
+        loop {
+            match self.inner.as_mut().unwrap().rx.recv().await {
+                Some(Ok(value_opt)) => return Ok(Some(value_opt)),
+                Some(Err(err)) => {
+                    if err.is_final() {
+                        if self.final_err.is_none() {
+                            self.final_err = Some(err);
+                        }
+                        continue;
+                    } else {
+                        return Err(err);
+                    }
+                }
+                None => match self.take_error() {
+                    Some(err) => return Err(err),
+                    None => return Ok(None),
+                },
+            }
         }
     }
 
     /// Polls to receive the next message on this channel.
     ///
-    /// This function returns `Poll::Ready(Ok(None))` when the channel sender has been dropped.
+    /// This function returns `Poll::Ready(Ok(None))` when all channel senders have been dropped.
+    ///
+    /// When a receive error occurs due to a connection failure and other senders are still
+    /// present, it is held back and returned after all other senders have been dropped or failed.
+    /// Use [error](Self::error) to check if such an error is present.
     #[inline]
     pub fn poll_recv(&mut self, cx: &mut Context<'_>) -> Poll<Result<Option<T>, RecvError>> {
-        match ready!(self.inner.as_mut().unwrap().rx.poll_recv(cx)) {
-            Some(Ok(value_opt)) => Poll::Ready(Ok(Some(value_opt))),
-            Some(Err(err)) => Poll::Ready(Err(err)),
-            None => Poll::Ready(Ok(None)),
+        loop {
+            match ready!(self.inner.as_mut().unwrap().rx.poll_recv(cx)) {
+                Some(Ok(value_opt)) => return Poll::Ready(Ok(Some(value_opt))),
+                Some(Err(err)) => {
+                    if err.is_final() {
+                        if self.final_err.is_none() {
+                            self.final_err = Some(err);
+                        }
+                        continue;
+                    } else {
+                        return Poll::Ready(Err(err));
+                    }
+                }
+                None => match self.take_error() {
+                    Some(err) => return Poll::Ready(Err(err)),
+                    None => return Poll::Ready(Ok(None)),
+                },
+            }
+        }
+    }
+
+    /// Tries to receive the next message on this channel, if one is immediately available.
+    ///
+    /// This function returns `Err(RecvError::Closed)` when all channel senders have been dropped
+    /// and `Err(RecvError::Empty)` if no value to receive is currently available.
+    ///
+    /// When a receive error occurs due to a connection failure and other senders are still
+    /// present, it is held back and returned after all other senders have been dropped or failed.
+    /// Use [error](Self::error) to check if such an error is present.
+    #[inline]
+    pub fn try_recv(&mut self) -> Result<T, TryRecvError> {
+        loop {
+            match self.inner.as_mut().unwrap().rx.try_recv() {
+                Ok(Ok(value_opt)) => return Ok(value_opt),
+                Ok(Err(err)) => {
+                    if err.is_final() {
+                        if self.final_err.is_none() {
+                            self.final_err = Some(err);
+                        }
+                        continue;
+                    } else {
+                        return Err(err.into());
+                    }
+                }
+                Err(tokio::sync::mpsc::error::TryRecvError::Disconnected) => match self.take_error() {
+                    Some(err) => return Err(err.into()),
+                    None => return Err(TryRecvError::Closed),
+                },
+                Err(tokio::sync::mpsc::error::TryRecvError::Empty) => return Err(TryRecvError::Empty),
+            }
         }
     }
 
@@ -136,9 +262,29 @@ impl<T, Codec, Buffer> Receiver<T, Codec, Buffer> {
     }
 
     /// Closes the receiving half of a channel without dropping it.
+    ///
+    /// This allows to process outstanding values while stopping the sender from
+    /// sending new values.
     #[inline]
     pub fn close(&mut self) {
-        let _ = self.inner.as_mut().unwrap().closed_tx.send(true);
+        let mut inner = self.inner.as_mut().unwrap();
+        let _ = inner.closed_tx.send(Some(ClosedReason::Closed));
+        inner.closed = true;
+    }
+
+    /// Returns the first error that occurred during receiving due to a connection failure,
+    /// but is being held back because other senders are still connected to this receiver.
+    ///
+    /// Use [take_error](Self::take_error) to clear it.
+    pub fn error(&self) -> &Option<RecvError> {
+        &self.final_err
+    }
+
+    /// Returns the held back error and clears it.
+    ///
+    /// See [recv](Self::recv) and [error](Self::error) for details.
+    pub fn take_error(&mut self) -> Option<RecvError> {
+        self.final_err.take()
     }
 
     /// Sets the codec that will be used when sending this receiver to a remote endpoint.
@@ -146,6 +292,7 @@ impl<T, Codec, Buffer> Receiver<T, Codec, Buffer> {
         Receiver {
             inner: self.inner.take(),
             successor_tx: Mutex::new(None),
+            final_err: self.final_err.clone(),
             _codec: PhantomData,
             _buffer: PhantomData,
         }
@@ -160,6 +307,7 @@ impl<T, Codec, Buffer> Receiver<T, Codec, Buffer> {
         Receiver {
             inner: self.inner.take(),
             successor_tx: Mutex::new(None),
+            final_err: self.final_err.clone(),
             _codec: PhantomData,
             _buffer: PhantomData,
         }
@@ -185,9 +333,13 @@ where
 
 impl<T, Codec, Buffer> Drop for Receiver<T, Codec, Buffer> {
     fn drop(&mut self) {
-        let mut successor_tx = self.successor_tx.lock().unwrap();
-        if let Some(successor_tx) = successor_tx.take() {
-            let _ = successor_tx.send(self.inner.take().unwrap());
+        if let Some(inner) = self.inner.take() {
+            let mut successor_tx = self.successor_tx.lock().unwrap();
+            if let Some(successor_tx) = successor_tx.take() {
+                let _ = successor_tx.send(inner);
+            } else if !inner.closed {
+                let _ = inner.closed_tx.send(Some(ClosedReason::Dropped));
+            }
         }
     }
 }
@@ -211,7 +363,8 @@ where
         let port = PortSerializer::connect(|connect| {
             async move {
                 // Receiver has been dropped after sending, so we receive its channels.
-                let ReceiverInner { mut rx, closed_tx, remote_send_err_tx } = match successor_rx.await {
+                let ReceiverInner { mut rx, closed_tx, remote_send_err_tx, closed: _ } = match successor_rx.await
+                {
                     Ok(inner) => inner,
                     Err(_) => return,
                 };
@@ -225,55 +378,18 @@ where
                     }
                 };
 
-                // Encode data using remote sender.
-                let mut remote_tx = base::Sender::<Result<T, RecvError>, Codec>::new(raw_tx);
-
-                // Process events.
-                let mut backchannel_active = true;
-                loop {
-                    tokio::select! {
-                        biased;
-
-                        // Back channel message from remote endpoint.
-                        backchannel_msg = raw_rx.recv(), if backchannel_active => {
-                            match backchannel_msg {
-                                Ok(Some(mut msg)) if msg.remaining() >= 1 => {
-                                    match msg.get_u8() {
-                                        BACKCHANNEL_MSG_CLOSE => {
-                                            let _ = closed_tx.send(true);
-                                        }
-                                        BACKCHANNEL_MSG_ERROR => {
-                                            let _ = remote_send_err_tx.send(Some(RemoteSendError::Forward));
-                                        }
-                                        _ => (),
-                                    }
-                                },
-                                _ => {
-                                    let _ = closed_tx.send(true);
-                                    backchannel_active = false;
-                                },
-                            }
-                        }
-
-                        // Data to send to remote endpoint.
-                        res_opt = rx.recv() => {
-                            let res = match res_opt {
-                                Some(res) => res,
-                                None => break,
-                            };
-
-                            if let Err(err) = remote_tx.send(res).await {
-                                let _ = remote_send_err_tx.send(Some(RemoteSendError::Send(err.kind)));
-                            }
-                        }
-                    }
-                }
+                send_impl!(T, rx, raw_tx, raw_rx, remote_send_err_tx, closed_tx);
             }
             .boxed()
         })?;
 
         // Encode chmux port number in transport type and serialize it.
-        let transported = TransportedReceiver::<T, Codec> { port, data: PhantomData, codec: PhantomData };
+        let transported = TransportedReceiver::<T, Codec> {
+            port,
+            data: PhantomData,
+            codec: PhantomData,
+            closed: self.inner.as_ref().unwrap().closed,
+        };
         transported.serialize(serializer)
     }
 }
@@ -293,11 +409,12 @@ where
         assert!(Buffer::size() > 0, "BUFFER must not be zero");
 
         // Get chmux port number from deserialized transport type.
-        let TransportedReceiver { port, .. } = TransportedReceiver::<T, Codec>::deserialize(deserializer)?;
+        let TransportedReceiver { port, closed, .. } =
+            TransportedReceiver::<T, Codec>::deserialize(deserializer)?;
 
         // Create channels.
         let (tx, rx) = tokio::sync::mpsc::channel(Buffer::size());
-        let (closed_tx, mut closed_rx) = tokio::sync::watch::channel(false);
+        let (closed_tx, mut closed_rx) = tokio::sync::watch::channel(None);
         let (remote_send_err_tx, mut remote_send_err_rx) = tokio::sync::watch::channel(None);
 
         PortDeserializer::accept(port, |local_port, request| {
@@ -311,51 +428,11 @@ where
                     }
                 };
 
-                // Decode received data using remote receiver.
-                let mut remote_rx = base::Receiver::<Result<T, RecvError>, Codec>::new(raw_rx);
-
-                // Process events.
-                let mut close_sent = false;
-                loop {
-                    tokio::select! {
-                        biased;
-
-                        // Channel closure requested locally.
-                        res = closed_rx.changed() => {
-                            match res {
-                                Ok(()) if *closed_rx.borrow() && !close_sent => {
-                                    let _ = raw_tx.send(vec![BACKCHANNEL_MSG_CLOSE].into()).await;
-                                    close_sent = true;
-                                }
-                                Ok(()) => (),
-                                Err(_) => break,
-                            }
-                        }
-
-                        // Notify remote endpoint of error.
-                        Ok(()) = remote_send_err_rx.changed() => {
-                            if remote_send_err_rx.borrow().as_ref().is_some() {
-                                let _ = raw_tx.send(vec![BACKCHANNEL_MSG_ERROR].into()).await;
-                            }
-                        }
-
-                        // Data received from remote endpoint.
-                        res = remote_rx.recv() => {
-                            let value = match res {
-                                Ok(Some(value)) => value,
-                                Ok(None) => break,
-                                Err(err) => Err(RecvError::RemoteReceive(err)),
-                            };
-                            if tx.send(value).await.is_err() {
-                                break;
-                            }
-                        }
-                    }
-                }
+                recv_impl!(T, tx, raw_tx, raw_rx, remote_send_err_rx, closed_rx);
             }
             .boxed()
         })?;
 
-        Ok(Self::new(rx, closed_tx, remote_send_err_tx))
+        Ok(Self::new(rx, closed_tx, closed, remote_send_err_tx))
     }
 }
